package com.bergerkiller.bukkit.nolagg;

import java.util.logging.Level;

import com.bergerkiller.bukkit.nolagg.sending.ChunkSendQueue;

import net.minecraft.server.Block;
import net.minecraft.server.Chunk;
import net.minecraft.server.WorldServer;

public class ChunkFixLightingOperation implements ThreadOperation {

	public static int processed = 0;
	
	private final int x, z;
	private final WorldServer world;
	private BufferedChunk data = null;
	public ChunkFixLightingOperation(org.bukkit.Chunk chunk) {
		this(ChunkHandler.getNative(chunk));
	}
	public ChunkFixLightingOperation(Chunk chunk) {
		this.world = (WorldServer) chunk.world;
		this.x = chunk.x;
		this.z = chunk.z;
	}
	
	@Override
	public void start() {
		Chunk chunk = this.getChunk(this.x, this.z);
		if (chunk != null) {
			this.data = new BufferedChunk(this.world, this.x, this.z);
			ChunkHandler.transferData(chunk, this.data);
		}
	}
	
	private Chunk getChunk(final int x, final int z) {
		return this.world.chunkProviderServer.chunks.get(x, z);
	}
	
	private int getLightLevel(int x, final int y, int z) {
		if (y <= 0 || y >= 127) return 0;
		if (x >= 0 && z >= 0 && x < 16 && z < 16) {
			return this.data.h.a(x, y, z);
		}
		
	    Chunk chunk = this.getChunk(this.data.x + (x >> 4), this.data.z + (z >> 4));
		if (chunk == null) return 0;
		x -= (chunk.x - this.data.x) << 4;
		z -= (chunk.z - this.data.z) << 4;
		return chunk.h.a(x, y, z);
	}
	private boolean fixBlockLighting(final int x, final int y, final int z) {
		final int typeid = this.data.getTypeId(x, y, z);
		if (!Block.o[typeid]) {
			final int factor = Math.max(1, Block.lightBlock[typeid]);
			final int light = this.data.h.a(x, y, z);
			//actual editing here
			int newlight = light + factor;
			newlight = Math.max(newlight, getLightLevel(x - 1, y, z));
			newlight = Math.max(newlight, getLightLevel(x + 1, y, z));
			newlight = Math.max(newlight, getLightLevel(x, y, z - 1));
			newlight = Math.max(newlight, getLightLevel(x, y, z + 1));
			newlight = Math.max(newlight, getLightLevel(x, y - 1, z));
			newlight = Math.max(newlight, getLightLevel(x, y + 1, z));
			newlight -= factor;
			//when exceeding in loops, just pick the highest value
			if (newlight > light) {
				this.data.h.a(x, y, z, newlight);
				return true;
			}
		}
		return false;
	}

	@Override
	public void run() {
		if (this.data == null) return;
		boolean haserror = true;
		int loops = -1;
		int x, y, z;
		while (haserror) {
			if (loops > 20) {
				NoLagg.log(Level.WARNING, "Failed to fix all lighting issues in chunk " + this.data.toString());
				break;
			}
			haserror = false;
			loops++;
			for (x = 0; x < 16; x++) {
				for (z = 0; z < 16; z++) {
					for (y = this.data.b(x,  z); y > 0; --y) {
						if (fixBlockLighting(x, y, z)) {
							haserror = true;
						}
					}
				}
			}
		}
	}

	@Override
	public void end() {
		if (this.data == null) return;
		//transfer block data
		Chunk chunk = this.getChunk(data.x, data.z);
		if (chunk != null) { 
			System.arraycopy(this.data.h.a, 0, chunk.h.a, 0, this.data.h.a.length);
			ChunkSendQueue.queueGlobal(chunk);
		}
		this.data.setFree();
		processed++;
	}

}
